from struct import unpack, pack
from impacket.structure import Structure
import binascii
import sys

# Keytab structure from http://www.ioplex.com/utilities/keytab.txt
  # keytab {
  #     uint16_t file_format_version;                    /* 0x502 */
  #     keytab_entry entries[*];
  # };

  # keytab_entry {
  #     int32_t size;
  #     uint16_t num_components;    /* sub 1 if version 0x501 */
  #     counted_octet_string realm;
  #     counted_octet_string components[num_components];
  #     uint32_t name_type;   /* not present if version 0x501 */
  #     uint32_t timestamp;
  #     uint8_t vno8;
  #     keyblock key;
  #     uint32_t vno; /* only present if >= 4 bytes left in entry */
  # };

  # counted_octet_string {
  #     uint16_t length;
  #     uint8_t data[length];
  # };

  # keyblock {
  #     uint16_t type;
  #     counted_octet_string;
  # };

class KeyTab(Structure):
    structure = (
        ('file_format_version','H=517'),
        ('keytab_entry', ':')
    )
    def fromString(self, data):
        self.entries = []
        Structure.fromString(self, data)
        data = self['keytab_entry']
        while len(data) != 0:
            ktentry = KeyTabEntry(data)

            data = data[len(ktentry.getData()):]
            self.entries.append(ktentry)

    def getData(self):
        self['keytab_entry'] = b''.join([entry.getData() for entry in self.entries])
        data = Structure.getData(self)
        return data

class OctetString(Structure):
    structure = (
        ('len', '>H-value'),
        ('value', ':')
    )

class KeyTabContentRest(Structure):
    structure = (
        ('name_type', '>I=1'),
        ('timestamp', '>I=0'),
        ('vno8', 'B=2'),
        ('keytype', '>H'),
        ('keylen', '>H-key'),
        ('key', ':')
    )

class KeyTabContent(Structure):
    structure = (
        ('num_components', '>h'),
        ('realmlen', '>h-realm'),
        ('realm', ':'),
        ('components', ':'),
        ('restdata',':')
    )
    def fromString(self, data):
        self.components = []
        Structure.fromString(self, data)
        data = self['components']
        for i in range(self['num_components']):
            ktentry = OctetString(data)

            data = data[ktentry['len']+2:]
            self.components.append(ktentry)
        self.restfields = KeyTabContentRest(data)

    def getData(self):
        self['num_components'] = len(self.components)
        # We modify the data field to be able to use the
        # parent class parsing
        self['components'] = b''.join([component.getData() for component in self.components])
        self['restdata'] = self.restfields.getData()
        data = Structure.getData(self)
        return data

class KeyTabEntry(Structure):
    structure = (
        ('size','>I-content'),
        ('content',':', KeyTabContent)
    )

# Add your own keys here!
# Keys are tuples in the form (keytype, 'hexencodedkey')
# Common keytypes for Windows:
# 23: RC4
# 18: AES-256
# 17: AES-128
# Wireshark takes any number of keys in the keytab, so feel free to add
# krbtgt keys, service keys, trust keys etc
keys = [
    # krbtgt
    ( 3, 'c2ea9d0e949d6797'),
    (18, '022c2172d9674697261c46726a24d99bf1af42b1993db25eb1040b0783d3933a'),
    (17, '31bcc9b60c6d6b1078055ca39294d86e'),

    # Horst Handshake as given by mimikatz
    ( 3, 'c783fb9ea4d5c1d6'),
    (18, 'c5bac6192ee8a76ab55e154a7c5e0993d3f90d3ab63bebbf6e86890b4d545407'),
    (17, 'cc2ac07c244a5e4cf016782e2dfd7036'),

    # Tim Tester as given by mimikatz
    ( 3, '8caba1c1648ff70e'),
    (18, 'cdc8c4465dfdf15d0334da9a23454a2819edbe0c21999d4c46d2055b93ec1b39'),
    (17, '4a58def5b2820f16aebc6c4315ff7ec3'),

    # Rainer Unsinn as given by mimikatz
    ( 3, 'b68a83adcba8ea10'),
    (18, '5ec630392ad5c7acabbbd91c2d66ca7099e33047118920a5a388589f3d221c43'),
    (17, 'd2dd040b5769a15328300bac0c5dfeba'),

    # administrator as given by mimikatz
    ( 3, '25e3c2166e38c8ab'),
    (18, '0fd95ffe4e3632b70f7535ae31d6685d3e52d799984e542cdeb54079e3202f20'),
    (17, 'dd473309d66b2eae8444754b6b2cd578'),

    # DC1$ as given by mimikatz
    ( 3, '79c70461d3ad08c2'),
    (18, 'afcb72b50e8ffd2a04cab4a3c6a48ed64076f5825ed8b448b5a5fd3a96660aa5'),
    (17, '2ff1d05f4a1227e3f2bd05b60c7b8e82'),

    # DC2$ as given by mimikatz
    ( 3, '76fe751f7f2a0b5e'),
    (18, '833f065cd00a8e069847add46d171654e2e778f0ab7213f13b8e22812c4684be'),
    (17, '9894aff69216b71205de0d383151d68a'),

    # LINUX-AES$ as given by mimikatz
    ( 3, '80687f1cd508ea7a'),
    (18, '60d968e62dc695199371e4c70c510898f5c74f44f0a4a74f428d886b4ec0aea6'),
    (17, 'b4e491719ddbad2a26b7faa36639751d')

]

nkt = KeyTab()
nkt.entries = []

for key in keys:
    ktcr = KeyTabContentRest()
    ktcr['keytype'] = key[0]
    ktcr['key'] = binascii.unhexlify(key[1])
    nktcontent = KeyTabContent()
    nktcontent.restfields = ktcr
    # The realm here doesn't matter for wireshark but does of course for a real keytab
    nktcontent['realm'] = b'EXCUSE-R-US.LOCAL'
    krbtgt = OctetString()
    krbtgt['value'] = 'krbtgt'
    nktcontent.components = [krbtgt]
    nktentry = KeyTabEntry()
    nktentry['content'] = nktcontent
    nkt.entries.append(nktentry)

data = nkt.getData()
if len(sys.argv) < 2:
    print('Usage: keytab.py <outputfile>')
    print('Keys should be written to the source manually')
else:
    with open(sys.argv[1], 'wb') as outfile:
        outfile.write(data)
